﻿//=============================================================================
// IRPP_MV_TitleMenuCommandEX.js
//=============================================================================

/*:
 * @plugindesc (※上部)タイトルやメニューコマンドをカスタマイズします。
 * @author イロスマRPG制作委員会
 *
 * @param Page1
 * @text コマンド名
 *
 * @param Title Option
 * @desc タイトルコマンドの「オプション」の名称です。
 * @default オプション
 * @parent Page1
 *
 * @param Title Shutdown
 * @desc タイトルコマンドの「シャットダウン」の名称です。
 * @default シャットダウン
 * @parent Page1
 *
 * @param GameEnd Shutdown
 * @desc ゲーム終了の「シャットダウン」の名称です。
 * @default シャットダウン
 * @parent Page1
 *
 * @param Page2
 * @text オプション
 *
 * @param Show Continue Title
 * @desc タイトル画面で「コンティニュー」を表示しますか？
 * @default 1
 * @type select
 * @option false
 * @value 0
 * @option true
 * @value 1
 * @parent Page2
 *
 * @param Show Option Title
 * @desc タイトル画面で「オプション」を表示しますか？
 * @default 1
 * @type select
 * @option false
 * @value 0
 * @option true
 * @value 1
 * @parent Page2
 *
 * @param Show Shutdown Title
 * @desc タイトル画面で「シャットダウン」を表示しますか？
 * ※ローカル環境以外の場合は非表示になります。
 * @default 1
 * @type select
 * @option false
 * @value 0
 * @option true
 * @value 1
 * @parent Page2
 *
 * @param Show Option Menu
 * @desc メニュー画面で「オプション」を表示しますか？
 * @default 1
 * @type select
 * @option false
 * @value 0
 * @option true
 * @value 1
 * @parent Page2
 *
 * @param Show GameEnd Menu
 * @desc メニュー画面で「ゲーム終了」を表示しますか？
 * @default 1
 * @type select
 * @option false
 * @value 0
 * @option true
 * @value 1
 * @parent Page2
 *
 * @param Show Shutdown GameEnd
 * @desc ゲーム終了で「シャットダウン」を表示しますか？
 * ※ローカル環境以外の場合は非表示になります。
 * @default 1
 * @type select
 * @option false
 * @value 0
 * @option true
 * @value 1
 * @parent Page2
 *
 * @help このプラグインにはプラグインコマンドがありません。
 * メニューコマンドのオプションの名称はデータベースで設定した物が適用されます。
 */

var Imported = Imported || {};
Imported.IRPP_MV_TitleMenuCommandEX = true;
(function() {
var Parameters = PluginManager.parameters('IRPP_MV_TitleMenuCommandEX');
var textTitleOption = String(Parameters['Title Option'] || 'Option');
var textTitleShutdown = String(Parameters['Title Shutdown'] || 'Shutdown');
var textGameEndShutdown = String(Parameters['GameEnd Shutdown'] || 'Shutdown');
var showContinueTitle = Number(Parameters['Show Continue Title'] || 1);
var showOptionTitle = Number(Parameters['Show Option Title'] || 1);
var showShutdownTitle = Number(Parameters['Show Shutdown Title'] || 1);
var showOptionMenu = Number(Parameters['Show Option Menu'] || 1);
var showGameEndMenu = Number(Parameters['Show GameEnd Menu'] || 1);
var showShutdownGameEnd = Number(Parameters['Show Shutdown GameEnd'] || 1);

Window_TitleCommand.prototype.makeCommandList = function() {
    this.addCommand(TextManager.newGame,   'newGame');
    if (showContinueTitle == 1) {
        this.addCommand(TextManager.continue_, 'continue', this.isContinueEnabled());
    }
    if (showOptionTitle == 1) {
        this.addCommand(textTitleOption,   'options');
    }
    if (showShutdownTitle == 1 && Utils.isNwjs()) {
        this.addCommand(textTitleShutdown,   'shutdown');
    }
};

Scene_Title.prototype.createCommandWindow = function() {
    this._commandWindow = new Window_TitleCommand();
    this._commandWindow.setHandler('newGame',  this.commandNewGame.bind(this));
    this._commandWindow.setHandler('continue', this.commandContinue.bind(this));
    this._commandWindow.setHandler('options',  this.commandOptions.bind(this));
    this._commandWindow.setHandler('shutdown',  this.commandShutdown.bind(this));
    this.addWindow(this._commandWindow);
};

Scene_Title.prototype.commandShutdown = function() {
    this._commandWindow.close();
    this.fadeOutAll();
    SceneManager.exit();
};

var _Window_MenuCommand_addOptionsCommand = Window_MenuCommand.prototype.addOptionsCommand;
Window_MenuCommand.prototype.addOptionsCommand = function() {
    if (showOptionMenu == 1) {
        _Window_MenuCommand_addOptionsCommand.call(this);
    }
};

var _Window_MenuCommand_addGameEndCommand = Window_MenuCommand.prototype.addGameEndCommand;
Window_MenuCommand.prototype.addGameEndCommand = function() {
    if (showGameEndMenu == 1) {
        _Window_MenuCommand_addGameEndCommand.call(this);
    }
};

Window_GameEnd.prototype.makeCommandList = function() {
    this.addCommand(TextManager.toTitle, 'toTitle');
    if (showShutdownGameEnd == 1 && Utils.isNwjs()) {
        this.addCommand(textGameEndShutdown,   'shutdown');
    }
    this.addCommand(TextManager.cancel,  'cancel');
};

Scene_GameEnd.prototype.createCommandWindow = function() {
    this._commandWindow = new Window_GameEnd();
    this._commandWindow.setHandler('toTitle',  this.commandToTitle.bind(this));
    this._commandWindow.setHandler('shutdown',  this.commandShutdown.bind(this));
    this._commandWindow.setHandler('cancel',   this.popScene.bind(this));
    this.addWindow(this._commandWindow);
};

Scene_GameEnd.prototype.commandShutdown = function() {
    this._commandWindow.close();
    this.fadeOutAll();
    SceneManager.exit();
};
})();